import numpy
import logging
import sys 
from apgl.predictors.LibSVM import LibSVM
from apgl.util.PathDefaults import PathDefaults
from apgl.util.Evaluator import Evaluator
from apgl.util.Util import Util
from apgl.util.FileLock import FileLock

"""
Let's run cross validation and model penalisation over the benchmark datasets from
Gunter Ratsch. All datasets are scaled to have unit variance and zero norm. 
"""
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

sampleSizes = numpy.array([50, 100, 200])
#sampleSizes = numpy.array([50])
foldsSet = numpy.arange(2, 11, 2)
#foldsSet = numpy.array([5])
dataDir = PathDefaults.getDataDir() + "modelPenalisation/benchmark/"
outputDir = PathDefaults.getOutputDir() + "modelPenalisation/"

datasetNames = []
datasetNames.append("banana")
datasetNames.append("breast-cancer")
datasetNames.append("diabetis")
datasetNames.append("flare-solar")
datasetNames.append("german")
datasetNames.append("heart")
datasetNames.append("image")
datasetNames.append("ringnorm")
datasetNames.append("splice")
datasetNames.append("thyroid")
datasetNames.append("titanic")
datasetNames.append("twonorm")
datasetNames.append("waveform")

numDatasets = len(datasetNames)
numMethods = 3
numParams = 2
numRealisations = numpy.ones(numDatasets, numpy.int)*100
numRealisations[6] = 20
numRealisations[8] = 20
#numRealisation = numpy.ones(numDatasets, numpy.int)

for i in range(len(datasetNames)):
    logging.debug("Learning using dataset " + datasetNames[i])
    outfileName = outputDir + datasetNames[i] + "Results"

    fileLock = FileLock(outfileName + ".npz")
    if not fileLock.isLocked() and not fileLock.fileExists():
        fileLock.lock()
        errors = numpy.zeros((numRealisations[i], len(sampleSizes), foldsSet.shape[0], numMethods))
        params = numpy.zeros((numRealisations[i], len(sampleSizes), foldsSet.shape[0], numMethods, numParams))

        for j in range(numRealisations[i]):
            Util.printIteration(j, 1, numRealisations[i], "Realisation: ")
            trainDatasetName = dataDir + datasetNames[i] + "/" + datasetNames[i] + "_train_data_" + str(j+1) + ".asc"
            trainLabelsName = dataDir + datasetNames[i] + "/" + datasetNames[i] + "_train_labels_" + str(j+1) + ".asc"
            testDatasetName = dataDir + datasetNames[i] + "/" + datasetNames[i] + "_test_data_" + str(j+1) + ".asc"
            testLabelsName = dataDir + datasetNames[i] + "/" + datasetNames[i] + "_test_labels_" + str(j+1) + ".asc"

            trainX = numpy.loadtxt(trainDatasetName, delimiter=None)
            trainY = numpy.loadtxt(trainLabelsName, delimiter=None)
            testX = numpy.loadtxt(testDatasetName, delimiter=None)
            testY = numpy.loadtxt(testLabelsName, delimiter=None)

            for k in range(sampleSizes.shape[0]):
                sampleSize = sampleSizes[k]
                for m in range(foldsSet.shape[0]):
                    folds = foldsSet[m]
                    logging.debug("Using sample size " + str(sampleSize) + " and " + str(folds) + " folds")
                    perm = numpy.random.permutation(trainX.shape[0])
                    trainInds = perm[0:sampleSize]
                    validX = trainX[trainInds, :]
                    validY = trainY[trainInds]

                    svm = LibSVM()

                    #Cross validation
                    #bestSVM = svm.vfcvRbf(validX, validY, folds)
                    logging.debug("Running V-fold cross validation")
                    bestSVM = svm.parallelVfcvRbf(validX, validY, folds)
                    predY = bestSVM.predict(testX)
                    errors[j, k, m, 0] = Evaluator.binaryError(testY, predY)
                    params[j, k, m, 0, :] = numpy.array([bestSVM.getC(), bestSVM.getKernelParams()])

                    #BIC penalisation
                    logging.debug("Running BIC penalisation")
                    Cv = float((folds-1) * numpy.log(validX.shape[0])/2)
                    bestSVM = svm.parallelVfPenRbf(validX, validY, folds, Cv)
                    predY = bestSVM.predict(testX)
                    errors[j, k, m, 1] = Evaluator.binaryError(testY, predY)
                    params[j, k, m, 1, :] = numpy.array([bestSVM.getC(), bestSVM.getKernelParams()])

                    #v fold penalisation
                    logging.debug("Running V-fold penalisation")
                    Cv = float(folds-1)
                    bestSVM = svm.parallelVfPenRbf(validX, validY, folds, Cv)
                    predY = bestSVM.predict(testX)
                    errors[j, k, m, 2] = Evaluator.binaryError(testY, predY)
                    params[j, k, m, 2, :] = numpy.array([bestSVM.getC(), bestSVM.getKernelParams()])

        meanErrors = numpy.mean(errors, 0)
        print(meanErrors)

        meanParams = numpy.mean(params, 0)
        print(meanParams)

        numpy.savez(outfileName, errors, params)
        logging.debug("Saved results as file " + outfileName + ".npz")
        fileLock.unlock()
    else:
        logging.debug("Results already computed")

logging.debug("All done!")